import hashlib
import logging
import traceback
from collections import Iterable
from functools import wraps

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import EmptyResultSet
from django.db.models.sql.compiler import SQLCompiler
from django.utils.six import wraps

from acmin.utils import attr, memorize

logger = logging.getLogger(__name__)

_cacheable_models = set()


def cacheable(cls):
    _cacheable_models.add(cls)
    return cls


@memorize
def _get_cache_tables():
    return [attr(model, "_meta.db_table").lower() for model in _cacheable_models]


_keywords = ['update', 'insert', 'delete', 'alter', 'create', 'drop']


def _table_names(sql):
    all_tables = _get_cache_tables()
    sql = sql.lower().replace("\t", " ")
    if any([True for keyword in _keywords if f"{keyword} " in sql]):
        return []

    return [name for name in all_tables if name in sql]


def clear(cls):
    if cls in _cacheable_models:
        cache.clear()


def patch():
    if attr(settings, "ACMIN_ENABLE_CACHE"):
        def unset_raw_connection(original):
            def inner(compiler, *args, **kwargs):
                compiler.connection.raw = False
                try:
                    return original(compiler, *args, **kwargs)
                finally:
                    compiler.connection.raw = True

            return inner

        def patch_compiler(original):
            @wraps(original)
            @unset_raw_connection
            def inner(compiler, *args, **kwargs):
                table_names = None
                try:
                    s = f"{compiler.as_sql()}"
                    table_names = _table_names(s)
                    md5 = hashlib.md5(s.encode()).hexdigest()
                    if table_names:
                        cached_result = cache.get(md5)
                        if cached_result is not None:
                            if attr(settings, "ACMIN_SHOW_CACHE_INFO"):
                                logger.info(f"acmin-cache:{cached_result}")
                            return cached_result
                except EmptyResultSet:
                    pass
                except:
                    logger.error(traceback.format_exc())

                result = original(compiler, *args, **kwargs)
                if table_names:
                    if result.__class__ not in {tuple, list, frozenset, set} and isinstance(result, Iterable):
                        result = list(result)
                    cache.set_many({md5: result})

                return result

            return inner

        SQLCompiler.execute_sql = patch_compiler(SQLCompiler.execute_sql)



from django import forms

from acmin.models import import_model
from acmin.utils import attrs, import_class, memorize


def get_dynamic_form(app_name, model, form_exclude=None):
    form_exclude = form_exclude or []
    if isinstance(model, str):
        model = import_model(app_name, model)

    '''
    class BaseForm(forms.ModelForm):
      class Meta:
        model=form_model
        fields=attrs.attr(model,"form_fields","__all__")
        exclude=attrs.attr(model,"form_exclude",[])

    return BaseForm
    '''
    return type("Dynamic%sForm" % model.__name__, (forms.ModelForm,), dict(
        Meta=type("Meta", (), dict(
            model=model,
            fields=attrs.attr(model, "form_fields", "__all__"),
            exclude=attrs.attr(model, "form_exclude", []) + form_exclude
        )),
        __module__=__name__
    ))


@memorize
def load_form(app_name, model):
    try:
        form = import_class("%s.forms.%sForm" % (app_name, model.__name__))
    except Exception:
        form = get_dynamic_form(app_name, model)
    return form



import json
import os
import platform

from django.core.serializers.json import DjangoJSONEncoder

from .decorators import memorize


def to_json(data):
    return json.dumps(data, cls=DjangoJSONEncoder)


def first(l, default=None):
    return l[0] if l else default


def last(l, default=None):
    return l[len(l) - 1] if l else default


def is_windows() -> bool:
    return platform.system() == "Windows"


@memorize
def get_ip():
    file = '/data/host'
    if os.path.exists(file):
        with open(file, 'r') as f:
            return f.readline().strip()

    return "127.0.0.1"


@memorize
def get_domain():
    file = '/data/domain'
    if os.path.exists(file):
        with open(file, 'r') as f:
            return f.readline().strip()
    return "localhost"


def null_to_emtpy(data):
    if isinstance(data, (list, tuple)):
        return [null_to_emtpy(i) for i in data]
    if isinstance(data, dict):
        return {key: null_to_emtpy(value) for key, value in data.items()}
    if data is None:
        return ""
    return data



def attr2(obj, name, default, show_display=False):
    result = None
    if isinstance(obj, dict):
        result = obj.get(name, default)
    elif isinstance(obj, (list, tuple)):
        try:
            result = obj[int(name)]
        except Exception:
            result = None
    elif hasattr(obj, name):
        if show_display:
            dispaly_method = 'get_' + name + '_display'
            if hasattr(obj, dispaly_method):
                method = getattr(obj, dispaly_method)
                result = method()
            else:
                result = getattr(obj, name)
        else:
            result = getattr(obj, name)

    return result


def attr(obj, attr_name, default=None, show_display=False):
    name = str(attr_name).strip()
    result = None
    if obj is not None:
        if "." not in name:
            result = attr2(obj, name, default, show_display)
        else:
            parts = name.split(".", 1)
            result = attr(attr(obj, parts[0]), parts[1])

    if result is None:
        result = default
    return result


get = attr


def set(obj, name, value):
    # obj.name = value
    setattr(obj, name, value)
    return obj


def display(obj, attr_name):
    result = attr(obj, attr_name, default="", show_display=True)
    if isinstance(result, bool):
        result = "是" if result else '否'

    return result



from django.urls import reverse
from django.views.generic.edit import UpdateView

from .form import AdminFormView


class AdminUpdateView(AdminFormView, UpdateView):
    success_message = "更新成功!"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        is_clone = self.request.GET.get("clone") is not None
        context['form_action'] = self.get_object().get_absolute_url()
        context["view_type"] = 'update'
        if is_clone:
            obj = context["object"]
            obj.id = ""
            context["form_action"] = reverse(f'{self.model.__name__}-create')
        return context



from acmin.utils import attr


class AccessMixin:
    from django.contrib.auth import REDIRECT_FIELD_NAME
    login_url = None
    permission_denied_message = ''
    raise_exception = False
    redirect_field_name = REDIRECT_FIELD_NAME

    def get_login_url(self):
        from django.core.exceptions import ImproperlyConfigured
        from django.conf import settings
        login_url = self.login_url or settings.LOGIN_URL
        if not login_url:
            raise ImproperlyConfigured(
                '{0} is missing the login_url attribute. Define {0}.login_url, settings.LOGIN_URL, or override '
                '{0}.get_login_url().'.format(self.__class__.__name__)
            )
        return str(login_url)

    def get_permission_denied_message(self):
        return self.permission_denied_message

    def get_redirect_field_name(self):
        return self.redirect_field_name

    def handle_no_permission(self):
        from django.contrib.auth.views import redirect_to_login
        from django.core.exceptions import PermissionDenied
        if self.raise_exception or self.request.user.is_authenticated:
            raise PermissionDenied(self.get_permission_denied_message())
        return redirect_to_login(self.request.get_full_path(), self.get_login_url(), self.get_redirect_field_name())

    def dispatch(self, request, *args, **kwargs):
        if request.user.is_authenticated and self.has_permisson():
            return super().dispatch(request, *args, **kwargs)
        else:
            return self.handle_no_permission()

    def has_permisson(self):
        return True


class StaticFilterMixin:
    def get_max_cls(self):
        return None

    def get_static_filter(self):
        return []


class ContextMixin(object):
    @property
    def user(self):
        return attr(self, "request.user")

    @property
    def is_creatable(self) -> bool:
        return self.model.creatable

    @property
    def is_cloneable(self) -> bool:
        return self.is_creatable

    @property
    def is_exportable(self) -> bool:
        return True

    @property
    def is_editable(self) -> bool:
        return self.model.editable

    @property
    def is_removable(self) -> bool:
        return self.model.removable

    @property
    def is_viewable(self) -> bool:  # 是否可以查看详情页,但不能编辑
        return True

    @property
    def is_operable(self) -> bool:
        return self.is_editable or self.is_removable

    @property
    def is_selectable(self) -> bool:
        return False

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            'model_name': self.model.__name__,
            'model_verbose_name': attr(self, 'model._meta.verbose_name'),
            'creatable': self.is_creatable,
            'cloneable': self.is_cloneable,
            'editable': self.is_editable,
            'viewable': self.is_editable,
            'removable': self.is_removable,
            'operable': self.is_operable,
            'exportable': self.is_exportable,
            "selectable": self.is_selectable,
        })
        return context



import json
from functools import wraps
from inspect import isfunction

from django.conf import settings
from django.http import HttpResponse
from django.urls import path as django_patb
from django.views.decorators.csrf import csrf_exempt

from acmin.utils import *

methods = {}


def get_urlpatterns():
    return [
        django_patb(r, get_view(func), name=f"{func.__module__}.{func.__qualname__}")
        for r, func in methods.items()
    ]


def admin_route(path):
    from django.conf import settings
    return route(path, f"{settings.ACMIN_ADMIN_PREFIX}/{settings.ACMIN_APP_NAME}" or "")


def route(path, prefix=""):
    def decorate(func):
        new_path = prefix + path
        if new_path.startswith("/"):
            new_path = new_path[1:]
        methods[new_path] = func

        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorate


def get_view(func):
    qualname = func.__qualname__
    module_name = func.__module__
    cls = None
    if "." in qualname:
        class_name, _ = qualname.split(".")
        cls = import_class("%s.%s" % (module_name, class_name))

        @csrf_exempt
        def view(request, *args, **kwargs):
            obj = cls()
            obj.request = request
            obj.args = args
            obj.kwargs = kwargs
            return func(obj, *args, **kwargs)
    else:
        def view(request, *args, **kwargs):
            return func(request, *args, **kwargs)

    return view


class Result(object):
    def __init__(self, status=1, data=None):
        self.status = status
        self.data = data


def to_json(data, fields):
    if 'QuerySet' in str(type(data)):  # data = obj_list 是个 QuerySet对象
        result = [to_json(obj, fields) for obj in data]  # 递归调用直到不存在QuerySet对象为止
    else:
        result = {}
        for name in fields:
            field = name
            if "|" in name:
                parts = name.split("|")  # 一、以 '|' 将字符串进行分割  得到一个列表，比如 'a|b'  ['a','b']
                name = parts[1]
                field = parts[0] + ".id"
            result[name] = attrs.attr(data, field)  # 将name作为key
    return result


class RouteMeta(type):
    def __new__(mcs, name, bases, attrs):
        cls = type.__new__(mcs, name, bases, attrs)

        for key, value in attrs.items():
            if isfunction(value):
                u = attr(value, "url")
                if u:
                    app_name = cls.__module__.split(".")[0]
                    prefix = f'/{cls._function_name}/{app_name}/{cls.__module__.split(".").pop()}'
                    setattr(cls, key, route(u, prefix)(value))

        return cls


class AcminView(metaclass=RouteMeta):

    def context(self):
        return {}

    def render(self, template_name: str, context: dict = None):
        updated_context = self.context() or {}
        updated_context.update(context or {})
        name = template_name[1:] if template_name.startswith("/") else template_name
        return render(self.request, name, updated_context)

    @staticmethod
    def json_response(dict_instance: dict):
        return json_response(dict_instance)
        # return HttpResponse(json.dumps(dict_instance, cls=CJsonEncoder), content_type="application/json")

    @classmethod
    def text_response(cls, text: str):
        return HttpResponse(text, content_type="text/html")

    @classmethod
    def error_json_message(cls, message: str = None):
        return cls.json_response({"status": 1, "message": message})

    @classmethod
    def ok_json_message(cls, message: str = None):
        return cls.json_response({"status": 0, "message": message})

    @classmethod
    def ok_data(cls, data=None):
        obj = {"status": 0}
        if data:
            obj['data'] = data
        return cls.json_response(obj)

    @classmethod
    def ok_list_json(cls, obj_list, include_fields=[], exclude_fields=[]):
        fields = include_fields
        if obj_list and not fields:  # obj_list 必须为真    fields必须为假    obj_list and not fields 此表达式才为真
            fields = [f for f in get_model_fields_without_relation(obj_list[0].__class__) if f not in exclude_fields]
        return cls.json_response({"status": 0, "data": to_json(obj_list, fields)})

    @classmethod
    def ok_object_json(cls, obj, include_fields=[], exclude_fields=[]):
        if obj:
            fields = include_fields or []
            if not fields:
                fields = [f for f in get_model_fields_without_relation(obj.__class__) if f not in exclude_fields]

            return cls.json_response({"status": 0, "data": to_json(obj, fields)})
        else:
            return cls.json_response({"status": 1})

    def int_param(self, param_name, default=0) -> int:
        return int_param(self.request, param_name, default)

    def bool_param(self, param_name, default=False) -> bool:
        return bool_param(self.request, param_name, default)

    def param(self, key, default=None) -> str:
        return param(self.request, key, default)

    def mac(self):
        return (self.param("mac", "") or self.param("imei", "")).upper()

    def file_json_response(self, file):
        return self.json_response(load_json(file))


class ApiView(AcminView):
    _function_name = "api"


class WebView(AcminView):
    _function_name = "web"


def url(param=None):
    if isfunction(param):
        param.url = "/" + param.__name__
        return get_wrapper(param)
    else:
        def decorate(func):
            if param is None:
                path = func.__name__
            else:
                path = param[1:] if param and param.startswith("/") else param
            func.url = "/" + path
            return get_wrapper(func)

        return decorate


def load_json(file):
    base = settings.BASE_DIR
    with open(f"{base}/data/{file}", 'r', encoding="UTF-8") as load_f:
        return json.load(load_f)



from django import template

register = template.Library()


def verbose(value, attr_name):
    fields = value.__class__._meta.fields
    verbose_name = [f.verbose_name for f in fields if f.name == attr_name].pop()
    if verbose_name:
        return verbose_name
    else:
        return ""


register.filter('verbose', verbose)



import logging
import threading

from django.conf import settings
from django.db.backends.utils import CursorWrapper
from django.db.models.sql.compiler import SQLInsertCompiler, SQLUpdateCompiler, SQLDeleteCompiler
from django.utils.six import wraps

from acmin.utils import attr

logger = logging.getLogger(__name__)


def patch():
    if attr(settings, 'ACMIN_SHOW_SQL'):
        def unset_raw_connection(original):
            def inner(compiler, *args, **kwargs):
                compiler.connection.raw = False
                try:
                    return original(compiler, *args, **kwargs)
                finally:
                    compiler.connection.raw = True

            return inner

        def patch_cursor(original):
            @wraps(original)
            def inner(cursor, sql, *args, **kwargs):
                try:
                    result = original(cursor, sql, *args, **kwargs)
                    return result
                finally:
                    logger.info(f"{threading.currentThread().ident} {sql},{args}")
                    pass

            return inner

        def patch_compiler(original):
            @wraps(original)
            @unset_raw_connection
            def inner(compiler, *args, **kwargs):
                try:
                    return original(compiler, *args, **kwargs)
                finally:
                    s = compiler.as_sql()
                    if type(s) is list and len(s) > 0:
                        s = s[0]
                    sql, params = s
                    logger.info(f"{threading.currentThread().ident} {sql}, {params}")

            return inner

        CursorWrapper.execute = patch_cursor(CursorWrapper.execute)
        CursorWrapper.executemany = patch_cursor(CursorWrapper.executemany)

        for c in (SQLInsertCompiler, SQLUpdateCompiler, SQLDeleteCompiler):  # SQLCompiler
            c.execute_sql = patch_compiler(c.execute_sql)



import ipaddress
import re

from django.core.exceptions import ValidationError

EVENT_ID_RE = re.compile(r'^[a-fA-F0-9]{32}$')


def validate_mac_address(value):
    if not re.match(r"^\s*([0-9a-fA-F]{2,2}:){5,5}[0-9a-fA-F]{2,2}\s*$", value):
        raise ValidationError('输入一个有效的 MAC 地址。')


def validate_ip_address(value):
    if not re.match(r"^\s*\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\s*$", value):
        raise ValidationError('输入一个有效的 IPv4 地址。')


def validate_ip(value, required=True):
    if not required and not value:
        return

    # will raise a ValueError
    ipaddress.ip_network(str(value), strict=False)
    return value


def is_float(var):
    try:
        float(var)
    except (TypeError, ValueError):
        return False
    return True


def is_event_id(value):
    try:
        return bool(EVENT_ID_RE.match(value))
    except TypeError:
        return False



import collections

import django.apps
from django.db.models.fields.related import ForeignKey

from . import attr, auto_repr, first, memorize


@memorize
def get_all_model_relations():
    class Relation:
        def __init__(self, model, attribute, verbose_name):
            self.attribute = attribute
            self.model = model
            self.verbose_name = verbose_name

    models = django.apps.apps.get_models()

    def init_model_relations():
        for model in models:
            fields = attr(model, '_meta.fields')
            for field in fields:
                remote_field = attr(field, "remote_field")
                if remote_field:
                    related_model = attr(remote_field, "model")
                    field = attr(remote_field, "field")
                    if type(field) is ForeignKey:
                        relations = attr(model, "relations", [])
                        relations.append(Relation(related_model, attr(field, "name"), attr(field, "verbose_name")))
                        setattr(model, "relations", relations)

    init_model_relations()

    def _get_attributes(model, name=None):
        relations = attr(model, "relations", [])
        names = []
        for relation in relations:
            new_name = f"{name}.{relation.attribute}" if name else relation.attribute
            new_names = _get_attributes(relation.model, new_name)
            if new_names:
                names += new_names
            else:
                names.append(new_name)

        return names

    def get_field_relation_model(model, attribute):
        _cls = model
        for name in attribute.split("."):
            field = attr(_cls, f"{name}.field")
            _cls = attr(field, f"remote_field.model")
        return _cls

    def _get_all(model):
        _result = []
        attributes = _get_attributes(model)
        if attributes:
            for attribute in attributes:
                names = attribute.split(".")
                for i in range(1, len(names) + 1):
                    sub_attribute = ".".join(names[0:i])
                    cls = get_field_relation_model(model, sub_attribute)
                    _result.append((sub_attribute, cls))
        return _result

    result = collections.OrderedDict()
    for model in models:
        relations = _get_all(model)
        groups, group = [], []
        last_attribute = None
        for relation in relations:
            if not group or relation[0].startswith(last_attribute):
                group.append(relation)
            elif group:
                groups.append(group)
                group = [relation]
            last_attribute = relation[0]
        if group:
            groups.append(group)
        result[model] = groups

    return result


def get_relation_group(model):
    return get_all_model_relations().get(model)


@memorize
def get_model_fields(cls) -> list:
    return attr(cls, '_meta.fields')


@memorize
def get_model_fields_without_relation(model):
    fields = model._meta.fields
    result = []
    for field in fields:
        name = field.name
        if 'ForeignKey' in str(field.__class__):
            name = "%s|%s_id" % (name, name)
        result.append(name)
    return result


@memorize
def get_many_to_many_fields(cls) -> list:
    return attr(cls, '_meta.many_to_many')


def get_model_field(cls, name):
    return first([f for f in attr(cls, '_meta.fields') if f.name == name])


@memorize
def get_model_field_names(cls) -> list:
    return [x.name for x in get_model_fields(cls)]


@memorize
def get_ancestor_attribute(child_cls, parent_cls, property_name=""):
    for name, x in get_parents(child_cls):
        name = name if not property_name else property_name + "." + name
        if x is child_cls or x is parent_cls:
            return name

        return get_ancestor_attribute(x, parent_cls, name)


@memorize
def get_ancestors(cls, max_cls=None):
    result = []
    if max_cls and not get_ancestor_attribute(cls, max_cls): return result
    if cls is not max_cls:
        foreign_fields = get_parents(cls)
        if foreign_fields:
            name, foreign_cls = foreign_fields[0]

            result.append((name, foreign_cls))
            if foreign_cls is not max_cls and foreign_cls is not cls:
                result += get_ancestors(foreign_cls, max_cls)
    return result


@memorize
def get_ancestors_names(cls, max_cls=None):
    return [name for (name, _) in get_ancestors(cls, max_cls)]


@memorize
def get_parents(cls) -> list:
    result = []
    for field in get_model_fields(cls):
        related_fields = attr(field, '_related_fields')
        if related_fields and isinstance(related_fields, list):
            item = related_fields[0]
            if isinstance(item, tuple):
                (i, _) = item
                if isinstance(i, ForeignKey):
                    result.append((field.name, i.related_model))
    return result


@auto_repr
class Field:
    def __init__(self, name, verbose, attribute_name, class_name=None, orderable=True, is_image=False):
        self.name = name
        self.verbose = verbose
        self.attribute_name = attribute_name
        self.class_name = class_name
        self.orderable = orderable
        self.is_image = is_image

# @memorize



from django.contrib.messages.views import SuccessMessageMixin
from django.http import HttpResponse
from django.views.generic.edit import DeleteView

from .mixins import AccessMixin


class AdminDeleteView(AccessMixin, SuccessMessageMixin, DeleteView):
    success_message = "删除成功!"

    def delete(self, request, *args, **kwargs):
        self.get_object().delete()
        return HttpResponse("ok")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["view_type"] = 'delete'
        return context



from django import template

register = template.Library()


@register.filter
def truncate(value, length=40):
    if value is not None and isinstance(value, str):
        if len(value) > length:
            return value[:length] + " ..."
    return value

# register.filter('truncate', truncate)



import logging

from acmin.forms import load_form
from acmin.utils import attrs, memorize, import_class

logger = logging.getLogger(__name__)


@memorize
def get_base_view_class(app_name, action):
    return import_class(f"{app_name}.views.Base{action.capitalize()}View")


def get_view(app_name, name, action):
    view_name = "%s%sView" % (name, action.capitalize())
    view = None
    try:
        view = import_class('%s.views.%s' % (app_name, view_name))
    except Exception:
        try:
            model_name = "%s.models.%s" % (app_name, name)
            entity = import_class(model_name)
            view = type("Dynamic%s" % view_name, (get_base_view_class(app_name, action),), dict(
                model=entity,
                __module__=f'{app_name}.{__name__}',

            ))
        except Exception as e:
            logger.error(e)

    if view and attrs.attr(view, 'form_class', None) is None:
        setattr(view, "form_class", load_form(app_name, view.model))

    return view.as_view()



from django.http import JsonResponse
from python_utils import converters

from . import null_to_emtpy


def param(request, key, default=None):
    if isinstance(key, list):
        return [param(request, item, default) for item in key]
    value = request.GET.get(key)
    if value is None:
        value = request.POST.get(key)
    if value is None:
        value = default
    return value


def int_param(request, param_name, default=0):
    return converters.to_int(param(request, param_name), default)


def bool_param(request, param_name ):
    value = param(request, param_name, "")
    return value.upper() in ["TRUE", "1", "OK"]


def json_response(dict_instance: dict):
    return JsonResponse(null_to_emtpy(dict_instance))



from django.urls import reverse
from django.views.generic.edit import CreateView

from .form import AdminFormView


class AdminCreateView(AdminFormView, CreateView):
    success_message = "创建成功!"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["form_action"] = reverse(f'{self.model.__name__}-create')
        context["view_type"] = 'create'

        return context



from django.apps import AppConfig


class AcminConfig(AppConfig):
    name = 'acmin'

    def ready(self):
        from acmin import sql
        from acmin import cache
        sql.patch()
        cache.patch()



import datetime
import gzip

import xlwt
from django.http import HttpResponse
from django.utils.http import urlquote
from django.utils.six import BytesIO

from acmin.utils import attr, display, decorators, get_model_fields, Field
from .list import AdminListView


@decorators.memorize
def get_export_fields(cls) -> list:
    names = attr(cls, "export_fields", attr(cls, "list_fields"))
    model_fields = get_model_fields(cls)
    if not names:
        export_fields = [f.name for f in model_fields]
        excludes = attr(cls, "export_exclude", attr(cls, "list_exclude", [])) + ['created', 'modified']
        names = [f for f in export_fields if f not in excludes]

    fields_dict = {f.name: f.verbose_name for f in model_fields}
    return [Field(name, fields_dict.pop(name, name), name) for name in names]


class AdminExportView(AdminListView):
    paginate_by = None

    def get_fields(self):
        return get_export_fields(self.model)

    def get(self, request, *args, **kwargs):
        return self.export_excel(self.get_queryset())

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["view_type"] = 'export'
        return context

    def export_excel(self, objects):
        fields = self.get_fields()
        model_verbose = attr(self.model, '_meta.verbose_name')
        wb = xlwt.Workbook(encoding='utf-8')
        sheet = wb.add_sheet('%s列表' % model_verbose)
        style_heading = xlwt.easyxf("""
            font:name Arial,colour_index white,bold on,height 0xA0;
            align:wrap off,vert center,horiz center;
            pattern:pattern solid,fore-colour 0x19;
            borders:left THIN,right THIN,top THIN,bottom THIN;
            """)
        style_body = xlwt.easyxf("""
                font:name Arial,bold off,height 0XA0;
                align:wrap on,vert center,horiz left;
                borders:left THIN,right THIN,top THIN,bottom THIN;
                """)
        # style_green = xlwt.easyxf(" pattern: pattern solid,fore-colour 0x11;")
        # style_red = xlwt.easyxf(" pattern: pattern solid,fore-colour 0x0A;")
        fmts = [
            'M/D/YY', 'D-MMM-YY', 'D-MMM', 'MMM-YY',
            'h:mm AM/PM', 'h:mm:ss AM/PM', 'h:mm', 'h:mm:ss', 'M/D/YY h:mm', 'mm:ss', '[h]:mm:ss', 'mm:ss.0',
        ]
        style_body.num_format_str = fmts[0]

        for column, field in enumerate(fields):
            sheet.write(0, column, str(field.verbose), style_heading)

        for row, obj in enumerate(objects, 1):
            for column, field in enumerate(fields, 0):
                sheet.write(row, column, str(display(obj, field.name)), style_body)

        buf = BytesIO()
        wb.save(buf)
        buf.seek(0)
        response = HttpResponse(gzip.compress(buf.getvalue()), content_type="application/vnd.ms-excel")
        filename = model_verbose + "-" + datetime.datetime.now().strftime('%Y%m%d%H%M%S')
        response['Content-Disposition'] = 'attachment;filename=%s.xls' % urlquote(filename)
        response['Content-Encoding'] = 'gzip'
        return response



from collections import OrderedDict

from django.contrib.messages.views import SuccessMessageMixin
from django.forms import ChoiceField

from acmin.utils import attr, get_ancestor_attribute, get_ancestors, get_ancestors_names
from .mixins import StaticFilterMixin,ContextMixin,AccessMixin


class AdminFormView(SuccessMessageMixin, StaticFilterMixin,ContextMixin,AccessMixin):
    def get_template_names(self):
        return [f"admin/{self.model.__name__}/form.html", 'base/form.html']

    def get_removed_fields(self):
        return []

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self._add_relation_choices(context)
        context["ancestors"] = self.get_ancestor_attribute_variables()
        context["model"] = self.model
        context["model_name"] = self.model.__name__

        return context

    def get_ancestor_attribute_variables(self):
        cls = self.model
        ancestors = list(reversed(get_ancestors(cls, self.get_max_cls())))
        length = len(ancestors)
        filters = self.get_static_filter()

        def to_dict(chain):
            name, cls = chain
            return dict(
                name=name,
                class_name=cls.__name__,
                filters={k: v for (c, d) in filters if c is cls for k, v in d.items()}
            )

        result = [dict(
            name=name,
            child=to_dict(ancestors[index + 1]),
            offsprings=[to_dict(c) for c in ancestors[index + 1:length]]
        ) for index, (name, cls) in enumerate(ancestors[0:length - 1])]
        return result

    def _add_relation_choices(self, context):
        obj = attr(context, "object")
        form = context["form"]
        choices = []
        max_cls = self.get_max_cls()
        chain_names = get_ancestors_names(self.model, max_cls)
        length = len(chain_names)
        chains = get_ancestors(self.model, max_cls)
        removed_fields = self.get_removed_fields()
        for index, (attribute, cls) in enumerate(chains):
            if attribute not in removed_fields:
                # print(attribute)
                queryset = cls.objects
                filters = self.get_static_filter()
                f = {k: v for c, f in filters if c is cls for k, v in f.items()}
                if f:
                    queryset = queryset.filter(**f)

                obj = attr(obj, attribute)
                if obj and index < length - 1:
                    parent_attribute_name = chain_names[index + 1]
                    parent_id = attr(obj, parent_attribute_name + ".id")
                    if parent_id:
                        f = {parent_attribute_name + "_id": parent_id}
                        queryset = queryset.filter(**f)
                else:
                    for c, f in filters:
                        relation = get_ancestor_attribute(cls, c)
                        if relation:
                            f = {relation.replace(".", "__") + "__" + key: value for key, value in f.items()}
                            queryset = queryset.filter(**f)

                choices.append((attribute, ChoiceField(
                    required=True if form.fields.pop(attribute, False) else False,
                    initial=attr(obj, "id"),
                    label=attr(cls, '_meta.verbose_name'),
                    choices=[('', '-----')] + [(e.id, str(e)) for e in queryset.all()]
                )))
        choices.reverse()

        fields = OrderedDict(choices)
        fields.update(form.fields)
        for f in removed_fields:
            fields.pop(f, None)
        form.fields = fields



from django import template

from acmin.utils import attr as u

register = template.Library()


@register.filter
def attr(value, attr_name):
    result = u(value, attr_name, default="")
    if isinstance(result, bool):
        result = "是" if result else '否'

    return result



from .attrs import attr, attr2, display, get, set
from .common import to_json, first, last, is_windows, get_ip, get_domain, null_to_emtpy
from .decorators import auto_repr, auto_str, memorize, get_wrapper
from .imports import import_class, import_sub_classes, import_submodules
from .models import (
    Field,
    get_ancestors,
    get_ancestors_names,
    get_model_fields,
    get_model_field,
    get_model_field_names,
    get_parents,
    get_ancestor_attribute,
    get_model_fields_without_relation

)
from .request import param, int_param, bool_param, json_response
from .string import is_empty, is_not_empty
from .template import render
from .validators import validate_mac_address



import time

import django.apps
import requests
from django.core.management.commands import loaddata


# python manage.py import host.com
class Command(loaddata.Command):
    def add_arguments(self, parser):
        super().add_arguments(parser)

    def handle(self, *args, **options):
        from django.conf import settings
        app_name = getattr(settings, "APP_NAME")
        if not settings.DEBUG: return
        for model in django.apps.apps.get_models():
            model.objects.all().delete()
        host = args[0]
        today = time.strftime("%Y-%m-%d")
        name = f"{today}.json.gz"
        url = f"https://{host}/backup/{app_name}/{name}"
        r = requests.get(url, stream=True)
        file = f"/tmp/{app_name}-{name}"
        f = open(file, "wb")
        for chunk in r.iter_content(chunk_size=512):
            if chunk: f.write(chunk)
        f.close()
        super().handle(*(file,), **options)



import operator
from collections import OrderedDict
from functools import reduce

from django.db.models import Q
from django.forms import ChoiceField, forms
from django.forms.models import model_to_dict
from django.http import JsonResponse
from django.views import generic
from django.views.generic.list import BaseListView

from acmin.utils import (
    attr, first, get_ancestor_attribute, get_ancestors, get_ancestors_names, get_model_field_names
)
from acmin.utils import models as model_util
from .mixins import StaticFilterMixin, ContextMixin, AccessMixin


class SearchMixin(BaseListView):
    def get_toolbar_search_params(self):
        groups = model_util.get_relation_group(self.model)
        if groups:
            relation_names = [x[0] for x in reduce(operator.add, groups)]
            return {k: v for k, v in self.request.GET.items() if k in relation_names and v}
        return {}

    def get_request_model_filters(self):
        return {k: v for k, v in self.request.GET.items() if k in get_model_field_names(self.model) and v}


class FuzzySearchMixin(BaseListView):
    def get_queryset(self):
        queryset = super().get_queryset()
        keyword = self.request.GET.get("sk")
        if keyword and len(keyword) > 0:
            search_fields = attr(self.model, "search_fields")
            if search_fields:
                q = Q()
                for field in search_fields:
                    q = q | Q(**{'%s__icontains' % field: keyword})
                queryset = queryset.filter(q)

        return queryset


class FilterForm(forms.Form):
    pass


class ToolbarSearchFormMixin(SearchMixin, StaticFilterMixin):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["toolbar_search_form"] = self.get_toolbar_search_form()
        names = [x for x in get_ancestors_names(self.model, self.get_max_cls())]
        context['hierarchy'] = {e: names[0:index] for index, e in enumerate(names)}
        return context

    def get_toolbar_search_form(self):
        choices = self.get_toolbar_search_fields()
        # print(choices)
        choices += self.get_toolbar_extra_search_choices()
        if choices:
            fields = OrderedDict(choices)
            form = FilterForm()
            form.fields = fields
            return form

    def get_toolbar_extra_search_choices(self):
        return []

    def _get_queryset_from_static_filters(self, cls):
        queryset = cls.objects
        filters = self.get_static_filter()
        if filters:
            for ancestor_cls, ancestor_filters in filters:
                attribute = get_ancestor_attribute(cls, ancestor_cls)
                if attribute:
                    f = {attribute.replace(".", "__") + "_" + key: value for key, value in ancestor_filters.items()}
                    queryset = cls.objects.filter(**f)
        return queryset

    def get_toolbar_search_fields(self):
        choices = []

        relation_groups = model_util.get_relation_group(self.model)
        params = self.get_toolbar_search_params()

        for relations in relation_groups:
            last_cls, last_name = None, None
            relations = list(reversed(relations))
            for name, cls in relations:
                queryset = None
                if last_name:
                    value = params.get(last_name, None)
                    if value:
                        attribute = get_ancestor_attribute(cls, last_cls)
                        queryset = cls.objects.filter(**{attribute.replace(".", "__") + "_id": value})
                else:
                    queryset = self._get_queryset_from_static_filters(cls)

                if queryset:
                    static_filter = {}
                    for (c, f) in self.get_static_filter():
                        if c is cls:
                            static_filter.update(f)
                    if static_filter:
                        queryset = queryset.filter(**static_filter)

                options = [(e.id, str(e)) for e in queryset.all()] if queryset else []
                if options:
                    label = attr(cls, '_meta.verbose_name')
                    choices.append((name, ChoiceField(
                        initial=params.get(name, ""),
                        label=label,
                        choices=[('', '选择%s' % label)] + options,
                    )))
                    last_cls, last_name = cls, name

        return choices


class OrderMixin(BaseListView):
    def get_queryset(self):
        queryset = super().get_queryset()
        sort = self.request.GET.get("sort")
        if sort:
            queryset = queryset.order_by(sort.replace('.', '__'))

        return queryset


class ToolbarSearchMixin(SearchMixin):
    def get_queryset(self):
        queryset = super().get_queryset()
        params = self.get_toolbar_search_params()
        for relations in model_util.get_relation_group(self.model):
            for attribute, cls in relations:
                value = params.get(attribute, None)
                if value:
                    queryset = queryset.filter(**{attribute.replace(".", "__") + "_id": value})
                    break

        model_filters = self.get_request_model_filters()
        if model_filters:
            queryset = queryset.filter(**model_filters)

        sort = self.request.GET.get("sort")
        if sort:
            queryset = queryset.order_by(sort.replace('.', '__'))

        return queryset


class JsonResponseMixin(BaseListView):
    def get(self, request, *args, **kwargs):
        if self.request.GET.get("format") == 'json':
            entities = [model_to_dict(item) for item in self.get_queryset()]
            return JsonResponse(entities, safe=False)

        return super().get(request, args, kwargs)


class AdminStaticListFilterMixin(StaticFilterMixin):
    def get_queryset(self):
        queryset = super().get_queryset()
        chains = get_ancestors(self.model)
        for cls, filters in self.get_static_filter():
            if self.model is cls:
                queryset = queryset.filter(**filters)
            else:
                index = first([index for index, (_, c) in enumerate(chains) if c is cls])
                if index is not None:
                    attribute = ".".join([name for name, _ in chains[0:index + 1]])
                    new_filters = {attribute.replace(".", "__") + "__" + k: v for k, v in filters.items()}
                    # print(new_filters)
                    queryset = queryset.filter(**new_filters)

        return queryset


class AdminListView(
    JsonResponseMixin,
    AdminStaticListFilterMixin,
    FuzzySearchMixin,
    ToolbarSearchFormMixin,
    ToolbarSearchMixin,
    OrderMixin,
    ContextMixin,
    AccessMixin,
    generic.ListView
):
    context_object_name = 'list'
    paginate_by = 30
    form_class = None

    def get_toolbar_search_fields(self):
        return super().get_toolbar_search_fields()

    def get_model_exclude_names(self):
        return attr(self.model, "list_exclude", [])

    def get_model_include_names(self):
        return attr(self.model, "list_fields")

    def get_model_list_fields(self):
        cls = self.model
        names = self.get_model_include_names()
        model_fields = model_util.get_model_fields(cls)
        field_dict = OrderedDict([(f.name, f) for f in model_fields])
        if names == '__all__' or not names:
            names = [f.name for f in model_fields]
        excludes = self.get_model_exclude_names()
        excludes = excludes + ['id', 'created', 'modified'] + [
            f.name for f in model_fields if f.related_model
        ]
        result = []
        for name in names:
            verbose_name = None
            orderable = True
            if isinstance(name, tuple):
                name, verbose_name = name
                orderable = False
            if name not in excludes:
                if not verbose_name:
                    verbose_name = field_dict[name].verbose_name if name in field_dict else name
                field = model_util.Field(name, verbose_name, name, None, orderable)
                result.append(field)

        return result

    def get_relation_fields(self):
        from acmin.utils import models
        fields = [model_util.Field(attribute_name.split(".").pop(), attr(clazz, "_meta.verbose_name"), attribute_name, clazz.__name__)
                  for relations in models.get_relation_group(self.model) for attribute_name, clazz in relations]

        fields.reverse()
        return fields

    def get_list_fields(self):
        return self.get_relation_fields() + self.get_model_list_fields()

    def get_template_names(self):
        return [f'admin/{self.model.__name__}/list.html', 'base/list.html']

    def get_context_data(self, **kwargs):
        from django.conf import settings
        context = super().get_context_data(**kwargs)
        context["view_type"] = 'list'
        context["media_url"] = attr(settings, "MEDIA_URL", "aaaa")
        context["image_height"] = getattr(settings, "ACMIN_IMAGE_HEIGHT", 80)
        context.update({"list_fields": self.get_list_fields(), })
        return context



import json
import linecache
import logging
import time
import tracemalloc
from collections import OrderedDict

logger = logging.getLogger(__name__)


def humanbytes(B):
    'Return the given bytes as a human friendly KB, MB, GB, or TB string'
    B = float(B)
    KB = float(1024)
    MB = float(KB ** 2)  # 1,048,576
    GB = float(KB ** 3)  # 1,073,741,824
    TB = float(KB ** 4)  # 1,099,511,627,776

    if B < KB:
        return '{0} {1}'.format(B, 'Bytes' if 0 == B > 1 else 'Byte')
    elif KB <= B < MB:
        return '{0:.2f} KB'.format(B / KB)
    elif MB <= B < GB:
        return '{0:.2f} MB'.format(B / MB)
    elif GB <= B < TB:
        return '{0:.2f} GB'.format(B / GB)
    elif TB <= B:
        return '{0:.2f} TB'.format(B / TB)


def display_top(snapshot, key_type='lineno', limit=5):
    snapshot = snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, "<unknown>"),
    ))
    top_stats = snapshot.statistics(key_type)

    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]  # tracemalloc.Frame
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            pass#print('    %s' % line)

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        #print("%s other: %.1f KiB" % (len(other), size / 1024))
    total = sum(stat.size for stat in top_stats)
    #print("Total allocated size: %.1f KiB" % (total / 1024))


def get_top(key, limit=5):
    snapshot = tracemalloc.take_snapshot()
    snapshot = snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap_external>"),
        tracemalloc.Filter(False, "<unknown>"),
    ))
    linenos = snapshot.statistics(key)
    # from tracemalloc import Statistic,Traceback
    result = []
    for index, stat in enumerate(linenos[:limit], 1):
        frame = stat.traceback[0]  # tracemalloc.Frame
        line = linecache.getline(frame.filename, frame.lineno).strip()
        obj = OrderedDict()
        obj["index"] = index
        obj["allocated"] = humanbytes(stat.size)
        obj["blocks"] = stat.count
        obj["filename"] = frame.filename
        obj["lineno"] = frame.lineno
        obj["code"] = line

        result.append(obj)
    return result


def show_trace():
    if not tracemalloc.is_tracing():
        logger.info("tracemalloc.start() begin")
        tracemalloc.start()
        time.sleep(1)

    logger.info(json.dumps({
        "memory_allocate": get_top('lineno'),
        "memory_block": get_top("traceback")
    }))



import threading
from functools import wraps


def get_wrapper(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return result

    return wrapper


def task(start=10, interval=60):
    def decorate(func):
        def proxy(*args, **kwargs):
            result = func(*args, **kwargs)
            threading.Timer(interval, proxy, args, kwargs).start()
            return result

        @wraps(proxy)
        def wrapper(*args, **kwargs):
            threading.Timer(start, proxy, args, kwargs).start()

        return wrapper

    return decorate


'''
cache_dict = {}
def memorize(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        key = f"[{func.__name__}-{args}-{kwargs}]"
        if key in cache_dict:
            # print("get from cache:"+key)
            return cache_dict[key]
        result = func(*args, **kwargs)
        cache_dict[key] = result
        return result

    return wrapper
'''


class memorize(dict):
    """
    Other version
    """

    def __init__(self, func):
        self.func = func

    def __call__(self, *args):
        return self[args]

    def __missing__(self, key):
        result = self[key] = self.func(*key)
        return result


def auto_str(cls):
    def f(self):
        return '%s(%s)' % (type(self).__name__, ', '.join('%s=%s' % item for item in vars(self).items()))

    cls.__str__ = f
    return cls


def auto_repr(cls):
    def f(self):
        return '%s(%s)' % (type(self).__name__, ', '.join('%s=%s' % item for item in vars(self).items()))

    cls.__repr__ = f
    return cls



#!/usr/bin/env python

import os

from setuptools import find_packages, setup

version = "0.1.18"

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'requirements.txt')) as f:
    required = f.read().splitlines()

setup(
    name='django-acmin',
    version=version,
    author='Hike Li',
    author_email='hikelee@gmail.com',
    url='https://github.com/hikelee/django-acmin',
    description='Admin and API template for start a django web',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Framework :: Django :: 2.1',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
    ],
    license='BSD',
    packages=find_packages(),
    install_requires=required,
    include_package_data=True,
    zip_safe=False,
)



from .admin import (
    AdminCreateView, AdminDeleteView, AdminExportView, AdminFormView, AdminListView, StaticFilterMixin, AdminUpdateView
)
from .admin import get_view as get_admin_view
from .view import ApiView, route, admin_route, WebView, url, get_urlpatterns



from django.conf import settings

def extra_context(request):
    return {
        'app_name': getattr(settings, "ACMIN_APP_NAME"),
        "function_name": getattr(settings, "ACMIN_FUNCTION_NAME"),
    }



import logging

from django import template
from django.conf import settings
from django.templatetags.static import StaticNode

from acmin.utils import get_domain, memorize

register = template.Library()
logger = logging.getLogger(__name__)


class ResourceNode(StaticNode):
    def url(self, context):
        path = self.path.resolve(context)
        prefix = get_prefix()
        if not path.startswith("http"):
            path = f"{prefix}/{path}"
        return path


@memorize
def get_prefix():
    return "/static" if settings.DEBUG else f"//{get_domain()}/static/{settings.APP_NAME}"


@register.tag('resource')
def do_static(parser, token):
    return ResourceNode.handle_token(parser, token)



from .base import get_base_view_class, get_view
from .create import AdminCreateView
from .delete import AdminDeleteView
from .export import AdminExportView
from .form import AdminFormView
from .list import AdminListView
from .mixins import StaticFilterMixin
from .update import AdminUpdateView



from .utils import import_class


class ModelMixin:
    def get_absolute_url(self):
        from django.urls import reverse
        cls = self.__class__
        return reverse(cls.__name__ + "-update", kwargs={"pk": self.pk})

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    @property
    def color(self):
        return "black"


def import_model(app_name, name):
    from django.apps import apps as django_apps
    if isinstance(name, str):
        result = None
        if "." in name:  # "hotel.Channel"
            result = django_apps.get_model(name)
        if not result:
            try:
                result = django_apps.get_model(app_name, name.lower())
            except Exception:
                pass
        if not result:
            try:
                result = import_class("%s.models.%s" % (app_name, name.capitalize()))
            except Exception:
                pass

        if result:
            return result
        raise Exception(f"Cannot import model by {app_name} {name}")



from django.test import TestCase

from acmin.utils import import_class


class Case(TestCase):
    def test_import_class(self):
        from acmin.forms import load_form
        #print(load_form("admin","BB"))
